/* CGV/solution/parts/cgvp02.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 02 <<<<<<<<<<<<<<<<<<<<<<*/

/*********************** halo routines ************************/
#ifndef serial
/*------------------------------
 * double* alloc_halo
 *   IN  int size: number of values (doubles) in the halo
 * RETURN:
 *   double* : the adress of the halo data
 * TASK:
 *   Allocates memory for the halo.
 * GLOBAL DATA:
 *   --
 * USED FUNCTIONS:
 *   APPL:
 *     safe_malloc
 *   C-lib:
 *     memset
 */
double* alloc_halo(int size)
{
#ifndef WITH_MEMOPS
    int i;
#endif
    double* halo = (double*) safe_malloc(sizeof(double) * size);
#ifdef WITH_MEMOPS
    memset(halo, 0, sizeof(double) * size);
#else
    for (i = 0; i < size; i++) {
        halo[i] = 0;
    }
#endif
    return halo;
}

/*------------------------------
 * void Init_Halo_Struct
 *
 * TASK:
 *   Initialize the halo structure.
 *
 * GLOBAL DATA:
 *   IN int rowsize : number of data entries in a row of the local chunk
 *   IN int colsize : number of data entries in a column of the local chunk
 *   IN/OUT halo_structure halo : the halo
 *   OUT double* send_scratch_buff: scratch buffer for sending data
 * USED FUNCTIONS:
 *   APPL:
 *     safe_malloc, alloc_halo
 *   MPI:
 *     MPI_Cart_shift
 */
void Init_Halo_Struct()
{
    int maxScratchBuffSize; 

    /* allocate sufficient memory for the send scratch buffer */
    maxScratchBuffSize = (rowsize > colsize)?rowsize:colsize;
    send_scratch_buff = (double*) safe_malloc(sizeof(double) * maxScratchBuffSize);

    /* determine neighbors (1-dim and 2-dims) */
    halo.north = alloc_halo(rowsize);
    halo.north_size = (rowsize);
    halo.south = alloc_halo(rowsize);
    halo.south_size = (rowsize);
    /* compute north_rank and south_rank */
          MPI_Cart_shift(comm_cart, 0, 1, &(halo.north_rank), &(halo.south_rank));

    halo.west = alloc_halo(colsize);
    halo.west_size = (colsize);
    halo.east = alloc_halo(colsize);
    halo.east_size = (colsize);
    /* compute west_rank and east_rank */
          MPI_Cart_shift(comm_cart, 1, 1, &(halo.west_rank), &(halo.east_rank)); 
}


/* halo communication: e.g. FROM_WEST_TO_EAST
 *
 *     *********************    *** *********************
 *     *                *  *    *W* *                   *
 *     *  WEST PROCESS  * -*--->*E* *  EAST PROCESS     *
 *     *                *  *    *S* *                   *
 *     *                *  *    *T* *                   *
 *     *********************    *** *********************
 *                               |
 *                               -----> WEST HALO
 */

/*------------------------------
 * void comm_vec_halo
 *
 * IN  vector_chunk v1: the vector with the data to send to other processes
 *
 * TASK:
 *   Send neccessary data to the neighbors
 *   and receive the needed data from the neighbors into
 *   the halo.
 *
 * USED GLOBAL DATA:
 *   IN OUT halo_structure halo              : the halo
 *   IN OUT double*        send_scratch_buff : scratch send buffer
 *   IN int rowsize : number of data entries in a row of the local chunk
 *   IN int colsize : number of data entries in a column of the local chunk
 *
* USED FUNCTIONS:
 *   MPI:
 *     MPI_Irecv, MPI_Send, MPI_Waitall
 */
/* message tags for the different communication directions */
#define TAG_FROM_NORTH_TO_SOUTH 101
#define TAG_FROM_EAST_TO_WEST   102
#define TAG_FROM_SOUTH_TO_NORTH 103
#define TAG_FROM_WEST_TO_EAST   104
void comm_vec_halo(vector_chunk v1)
{
    int i;
    MPI_Request req_array[4];
    MPI_Status  status_array[4];
    
    /* make a non-blocking receive */

    /* Receive halo (non blocking). If current rank has no neighbors, */
    /* the neighbor rank is MPI_PROC_NULL and nothing is sent. */
    /* receive north halo */
    MPI_Irecv(halo.north, halo.north_size, MPI_DOUBLE, halo.north_rank,
              TAG_FROM_NORTH_TO_SOUTH, comm_cart, &req_array[0]);
    /* receive east halo */
          MPI_Irecv(halo.east, halo.east_size, MPI_DOUBLE, halo.east_rank,
                    TAG_FROM_EAST_TO_WEST, comm_cart, &req_array[1]);
    /* receive south halo */
          MPI_Irecv(halo.south, halo.south_size, MPI_DOUBLE, halo.south_rank,
                    TAG_FROM_SOUTH_TO_NORTH, comm_cart, &req_array[2]);
    /* receive west halo */
          MPI_Irecv(halo.west, halo.west_size, MPI_DOUBLE, halo.west_rank,
                    TAG_FROM_WEST_TO_EAST, comm_cart, &req_array[3]);
    
    /* send to north neighbor (blocking) */
    MPI_Send(&v1[SIDX(0, 0)], halo.north_size, MPI_DOUBLE, halo.north_rank,
             TAG_FROM_SOUTH_TO_NORTH, comm_cart);

    /* do we have a east neighbor? */
          if (halo.east_rank != MPI_PROC_NULL) {
        /* fill scratch send buffer with last column of the local chunk */
              for (i = 0; i < halo.east_size; i++) {
                  send_scratch_buff[i] = v1[SIDX(i, (rowsize - 1))];
              }
          }
    /* send to east neighbor (blocking) */
    MPI_Send(send_scratch_buff, halo.east_size, MPI_DOUBLE, halo.east_rank,
             TAG_FROM_WEST_TO_EAST, comm_cart);

    /* send to south neighbor (blocking) */
          MPI_Send(&v1[SIDX((colsize - 1), 0)], halo.south_size, MPI_DOUBLE, halo.south_rank,
                   TAG_FROM_NORTH_TO_SOUTH, comm_cart);
    
    /* do we have a west neighbor? */
          if (halo.west_rank != MPI_PROC_NULL) {
        /* fill scratch send buffer with first column of the local chunk */
              for (i = 0; i < halo.west_size; i++) {
                  send_scratch_buff[i] = v1[SIDX(i, 0)];
              }
          }
    /* send to west neighbor (blocking) */
          MPI_Send(send_scratch_buff, halo.west_size, MPI_DOUBLE, halo.west_rank,
                   TAG_FROM_EAST_TO_WEST, comm_cart);

    /* wait for all halo info */
          MPI_Waitall(4, req_array, status_array);
}
#endif
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 02 <<<<<<<<<<<<<<<<<<<<<<*/
