/* CGV/solution/parts/cgvp03.c */

/****************************************************************
 *                                                              *
 * This file has been written as a sample solution to an        *
 * exercise in a course given at the High Performance           *
 * Computing Centre Stuttgart (HLRS).                           *
 * It is made freely available with the understanding that      *
 * every copy of this file must include this header and that    *
 * HLRS takes no responsibility for the use of the enclosed     *
 * teaching material.                                           *
 *                                                              *
 * Authors: Dr. Rolf Rabenseifner                               *
 *          Stephan Scheiderer                                  *
 *                                                              *
 * Contact: rabenseifner@hlrs.de                                *
 *                                                              *
 ****************************************************************/
                                                           /*EOH*/
/*>>>>>>>>>>>>>>>>>>>>>> begin of task 03 <<<<<<<<<<<<<<<<<<<<<<*/

/********************** matrix-vector-multiply ****************************/

/* OVERVIEW:
 * 1) Calculation of the "inner values"
 * 2) Calculation of the "border values" of the first row
 * 3) Calculation of the "border values" of the last row
 * 4) Calculation of the "border values" of the first column
 * 5) Calculation of the "border values" of the last column
 * 6) Calculation of the "corner values"
 */

/*------------------------------
 * mult_A_vec
 *      IN  vector_chunk v1 : vector to multiply A with
 *      OUT vector_chunk v2 : output vector
 *
 * TASK: Calculates matrix vector product v2 = A * v1
 *
 * USED GLOBAL DATA: 
 *   IN int start_i   : start row of the physical area this process is to calculate
 *   IN int end1_i    : end row + 1 of the physical area this process is to calculate
 *   IN int start_j   : start column of the physical area this process is to calculate
 *   IN int end1_j    : end column + 1 of the physical area this process is to calculate
 *   IN int rowsize   : number of data entries in a row of the local chunk
 *   IN int colsize   : number of data entries in a column of the local chunk
 *   IN halo_structure halo : the halo data from the neighbors
 *
 * USED FUNCTIONS:
 *   C-lib:
 *     printf
 */
void mult_A_vec(vector_chunk restrict v1, vector_chunk restrict /* out */ v2) 
{
    int i, j;

    /************************************************************************
     * 1) Calculation of the "inner values"                                 *
     ************************************************************************/

    /* first we calculate the "inner" values that do not access the boundaries */
    /* nor the halo */
    for (i = 1; i < (colsize - 1); i++) {
        for (j = 1; j < (rowsize - 1); j++) {
            v2[SIDX(i, j)] = -1 * v1[SIDX((i - 1), j)] +
                             -1 * v1[SIDX(i, (j - 1))] +    
                              4 * v1[SIDX(i, j)] +
                             -1 * v1[SIDX(i, (j + 1))] +
                             -1 * v1[SIDX((i + 1), j)];
        }
    }
    /* for the MFLOPS calculation */
    flops += (rowsize - 2) * (colsize - 2) * 9;

    /************************************************************************
     * 2) Calculation of the "border values" of the first row               *
     ************************************************************************/

    /* second the values in the first row of the local chunk */
    /* not taking the corner values at (start_i, start_j) and (start_i, end1_j - 1) */ 

    /* there are 2 different cases where we have to calculate the values differently */

    /* case A1: boundaries in the north, values in the south -> first row in the physical problem, */
    /*          local chunk has more than one row (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    if (start_i == 0) {
        for (j = 1; j < (rowsize - 1); j++) {
            /* i is set to 0 and the boundaries are taken into account */
            v2[SIDX(0, j)] = -1 * v1[SIDX(0, (j - 1))] +  
                              4 * v1[SIDX(0, j)] +
                             -1 * v1[SIDX(0, (j + 1))] +
                             -1 * v1[SIDX(1, j)];
        }
        /* for the MFLOPS calculation */
        flops += (rowsize - 2) * 7;
    }
#ifndef serial
    /* case A2: halo in the north, values in the south */
    /*          -> somewhere in the middle of the physical problem, */
    /*          local chunk has more than one row (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    else if (start_i > 0) {
        for (j = 1; j < (rowsize - 1); j++) {
            /* i is set to 0 (start_i - start_i) and the halo is taken into account */
            v2[SIDX(0, j)] = -1 * halo.north[j] +
                             -1 * v1[SIDX(0, (j - 1))] +  
                              4 * v1[SIDX(0, j)] +
                             -1 * v1[SIDX(0, (j + 1))] +
                             -1 * v1[SIDX(1, j)];
        }
        /* for the MFLOPS calculation */
        flops += (rowsize - 2) * 9;
    }
#endif


    /************************************************************************
     * 3) Calculation of the "border values" of the last row                *
     ************************************************************************/

    /* third the values in the last row of the local chunk */
    /* not taking the corner values at (end1_i - 1, start_j) */
    /* and (end1_i - 1, end1_j - 1) */ 

    /* there are 2 different cases where we have to calculate the values differently */

    /* case B1: values in the north, boundary in the south */
    /*          -> last row of the physical problem, */
    /*          local chunk has more than one row (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    if (end1_i == m) {
        for (j = 1; j < (rowsize - 1); j++) {
            /* i is set to (colsize - 1) and the boundary is taken into account */
            v2[SIDX((colsize - 1), j)] = -1 * v1[SIDX(((colsize - 1) - 1), j)] +
                                         -1 * v1[SIDX((colsize - 1), (j - 1))] +  
                                          4 * v1[SIDX((colsize - 1), j)] +
                                         -1 * v1[SIDX((colsize - 1), (j + 1))];
        }
        /* for the MFLOPS calculation */
        flops += (rowsize - 2) * 7;
    }
#ifndef serial
    /* case B2: values in the north, halo in the south */
    /*          -> in the middle of the physical problem, */
    /*          local chunk has more than one row (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    else if (end1_i != m) {
        for (j = 1; j < (rowsize - 1); j++) {
            /* i is set to (colsize - 1) and the halo is taken into account */
            v2[SIDX((colsize - 1), j)] = -1 * v1[SIDX(((colsize - 1) - 1), j)] +
                                         -1 * v1[SIDX((colsize - 1), (j - 1))] +  
                                          4 * v1[SIDX((colsize - 1), j)] +
                                         -1 * v1[SIDX((colsize - 1), (j + 1))] +
                                         -1 * halo.south[j];
        }
        /* for the MFLOPS calculation */
        flops += (rowsize - 2) * 9;
    }
#endif

    /************************************************************************
     * 4) Calculation of the "border values" of the first column            *
     ************************************************************************/

    /* fourth the values in the first column of the local chunk */
    /* not taking the corner values at (start_i, start_j) and (end1_i - 1, start_j) */ 

    /* there are 2 different cases where we have to calculate the values differently */
    
    /* case C1: boundary in the west, values in the east */
    /*          -> first column of the physical problem, */
    /*          local chunk has more than one column (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    if (start_j == 0) {
        for (i = 1; i < (colsize - 1); i++) {
            /* j is set to 0 and the boundary is taken into account */
            v2[SIDX(i, 0)] = -1 * v1[SIDX((i - 1), 0)] +
                              4 * v1[SIDX(i, 0)] +
                             -1 * v1[SIDX(i, 1)] +
                             -1 * v1[SIDX((i + 1), 0)];
        }
        /* for the MFLOPS calculation */
        flops += (colsize - 2) * 7;
    }
#ifndef serial
    /* case C2: halo in the west, values in the east */
    /*          -> in the middle of the physical problem, */
    /*          local chunk has more than one column (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    else if (start_j > 0) {
        for (i = 1; i < (colsize - 1); i++) {
            /* j is set to 0 and the halo is taken into account */
            v2[SIDX(i, 0)] = -1 * v1[SIDX((i - 1), 0)] +
                             -1 * halo.west[i] + 
                              4 * v1[SIDX(i, 0)] +
                             -1 * v1[SIDX(i, 1)] +
                             -1 * v1[SIDX((i + 1), 0)];
        }
        /* for the MFLOPS calculation */
        flops += (colsize - 2) * 9;
    }
#endif

    /************************************************************************
     * 5) Calculation of the "border values" of the last column             *
     ************************************************************************/

    /* fifth the values in the last column of the local chunk */
    /* not taking the corner values at (start_i, end1_j - 1) and (end1_i - 1, end1_j) */ 

    /* there are 2 different cases where we have to calculate the values differently */

    /* case D1: values in the west, boundary in the east */
    /*          -> last column of the physical problem, */
    /*          local chunk has more than one column (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    if (end1_j == n) {
        for (i = 1; i < (colsize - 1); i++) {
            /* j is set to (rowsize - 1) and the boundary is taken into account */
            v2[SIDX(i, (rowsize - 1))] = -1 * v1[SIDX((i - 1), (rowsize - 1))] +
                                         -1 * v1[SIDX(i, ((rowsize - 1) - 1))] + 
                                          4 * v1[SIDX(i, (rowsize - 1))] +
                                         -1 * v1[SIDX((i + 1), (rowsize - 1))];
        }
        /* for the MFLOPS calculation */
        flops += (colsize - 2) * 7;
    }
#ifndef serial
    /* case D2: values in the west, halo in the east */
    /*          local chunk has more than one column (we assure that each chunk has an extent */
    /*          of at least 2x2) */
    else if (end1_j != n) {
        for (i = 1; i < (colsize - 1); i++) {
            /* j is set to (rowsize - 1) and the halo is taken into account */
            v2[SIDX(i, (rowsize - 1))] = -1 * v1[SIDX((i - 1), (rowsize - 1))] +
                                         -1 * v1[SIDX(i, ((rowsize - 1) - 1))] + 
                                          4 * v1[SIDX(i, (rowsize - 1))] +
                                         -1 * halo.east[i] +
                                         -1 * v1[SIDX((i + 1), (rowsize - 1))];
        }
        /* for the MFLOPS calculation */
        flops += (colsize - 2) * 9;
    }
#endif

    /************************************************************************
     * 6) Calculation of the "corner values"                                *
     ************************************************************************/
    
    /* at last the values in the corners: (start_i, start_j), */
    /* (start_i, end1_j - 1), (end1_i - 1, start_j), (end1_i - 1, end1_j - 1) */
   
    /* we have to calculate all corner values, because we assure that each */
    /* chunk has at least an extent of 2x2 */
    /* For each corner value, we have different cases due to halo or boundary */
    /* neighborship */ 

    /* corner value at (start_i, start_j)*/
    /* must always be calculated */
    v2[SIDX(0, 0)] = 4 * v1[SIDX(0, 0)];

#ifndef serial
    /* is the corner value at the north boundary? */
    if (start_i > 0) {
        /* No -> halo in the north! */
        v2[SIDX(0, 0)] += -1 * halo.north[0];
    }
    /* is the corner value at the west boundary? */
    if (start_j > 0) {
        /* No -> halo in the west! */
        v2[SIDX(0, 0)] += -1 * halo.west[0];
    }
#endif
    /* we always have a value in the east, because we assure that each chunk has */
    /* at least 2x2 extent */
    v2[SIDX(0, 0)] += -1 * v1[SIDX(0, 1)];
    /* we always have a value in the south, because we assure that each chunk has */
    /* at least 2x2 extent */
    v2[SIDX(0, 0)] += -1 * v1[SIDX(1, 0)];

    /* corner value at (start_i, end1_j - 1) */
    /* must always be calculated */
    v2[SIDX(0, (rowsize - 1))] = -1 * v1[SIDX(0, ((rowsize - 1) - 1))] +
                                  4 * v1[SIDX(0, (rowsize - 1))];
#ifndef serial
    /* are we at the north boundary? */
    if (start_i != 0) {
        /* No -> halo in the north */
        v2[SIDX(0, (rowsize - 1))] += -1 * halo.north[(rowsize - 1)];
    }
    /* are we at the east boundary? */
    if (end1_j != n) {
        /* No -> halo in the east */
        v2[SIDX(0, (rowsize - 1))] += -1 * halo.east[0];
    }
#endif
    /* we always have a value in the south, because we assure that each chunk has */
    /* at least 2x2 extent */
    v2[SIDX(0, (rowsize - 1))] += -1 * v1[SIDX(1, (rowsize - 1))];

    /* corner value at (end1_i - 1, start_j) */
    /* must always be calculated */
    v2[SIDX((colsize - 1), 0)] = -1 * v1[SIDX(((colsize - 1) - 1), 0)] +
                                  4 * v1[SIDX((colsize - 1), 0)];

#ifndef serial
    /* are we at the west boundary? */
    if (start_j != 0) {
        /* No -> halo in the west */
        v2[SIDX((colsize - 1), 0)] += -1 * halo.west[(colsize - 1)];
    }
    /* are we at the south boundary? */
    if (end1_i != m) {
        /* No -> halo in the south */
        v2[SIDX((colsize - 1), 0)] += -1 * halo.south[0];
    }
#endif
    /* we always have a value in the east, because we assure that each chunk has */
    /* at least 2x2 extent */
    v2[SIDX((colsize - 1), 0)] += -1 * v1[SIDX((colsize - 1), 1)];

    /* corner value at (end1_i - 1, end1_j - 1) */
    /* must always be calculated */
    v2[SIDX((colsize - 1), (rowsize - 1))] = -1 * v1[SIDX((colsize - 1), ((rowsize - 1) - 1))] +
                                             -1 * v1[SIDX(((colsize - 1) - 1), (rowsize - 1))] +
                                              4 * v1[SIDX((colsize - 1), (rowsize - 1))];
#ifndef serial
    /* are we at the east boundary? */
    if (end1_j != n) {
        /* No -> halo in the east! */
        v2[SIDX((colsize - 1), (rowsize - 1))] += -1 * halo.east[(colsize - 1)];
    }
    /* are we at the south boundary? */
    if (end1_i != m) {
        /* No -> halo in the south! */
        v2[SIDX((colsize - 1), (rowsize - 1))] += -1 * halo.south[(rowsize - 1)];
    }
#endif
    /* MFLOPS for the corner values are neglected */
}
/*>>>>>>>>>>>>>>>>>>>>>> -end- of task 03 <<<<<<<<<<<<<<<<<<<<<<*/
